/*
** $PROJECT: template sub datatype class for HTDS
**
** $VER: classbase.c 40.2 (09.11.96)
**
** $AUTHOR: Stefan Ruppert , Windthorststrasse 5 , 65439 Floersheim , GERMANY
**
** $COPYRIGHT: (C) Copyright 1996 by Stefan Ruppert, All Rights Reserved !
**
** $EMAIL: Stefan.Ruppert@informatik.fh-wiesbaden.de
**
** $HISTORY:
**
** 09.11.96 : 040.002 : changed to new ClassLibrary structure fromthe DevCD
** 15.09.96 : 040.001 : + now uses semaphore for shared data as in AIFFdtc
**                        from Olaf Barthel
**                      + OpenLibrary() now in LibOpen() (suggestion from
**                        Stefan Becker)
**                      + added ROMTag and removed last asm file.
*/

/*FS*//* ------------------------------- includes ------------------------------- */

#include "classbase.h"

#include <exec/initializers.h>
#include <exec/resident.h>

#ifdef __SASC
#define ALIGNED __aligned
#else
#define ALIGNED
#endif
/*FE*/

/*FS*//* ------------------------------ prototypes ------------------------------ */

LibCall struct Library *LibInit (REGD0 struct ClassBase *cb, REGA0 BPTR seglist, REGA6 struct Library * sysbase);
LibCall LONG LibOpen (REGA6 struct ClassBase *cb);
LibCall LONG LibClose (REGA6 struct ClassBase *cb);
LibCall LONG LibExpunge (REGA6 struct ClassBase *cb);
LibCall LONG LibReserved (REGA6 struct ClassBase *cb);
LibCall Class *ObtainEngine (REGA6 struct ClassBase *cb);

static Class *initClass(struct ClassBase *cb);
static BOOL OpenLibraries(struct ClassBase *cb);
static void CloseLibraries(struct ClassBase *cb);

extern ClassCall ULONG dispatcher(REGA0 Class *cl,REGA2 Object *obj,REGA1 Msg msg);

#ifdef USERCLASSBASEINIT
extern BOOL UserClassBaseInit(struct ClassBase *cb);
extern void UserClassBaseClean(struct ClassBase *cb);
#endif
/*FE*/

/*FS*//* --------------------------------- data --------------------------------- */

static const ALIGNED UBYTE LibName[] = NAME;
static const ALIGNED UBYTE LibId[]   = VSTRING;

const APTR LibFuncTable[] =
{
   LibOpen,
   LibClose,
   LibExpunge,
   LibReserved,
   ObtainEngine,

   USERCLASSBASEFUNCTABLE

   (APTR) -1
};

const ALIGNED ULONG LibInitTable[4] = {
   (ULONG) sizeof(struct ClassBase),
   (ULONG) LibFuncTable,
   (ULONG) NULL,
   (ULONG) LibInit
   };

ALIGNED struct Resident romtag =
{
   RTC_MATCHWORD,
   &romtag,
   (&romtag + 1),
   RTF_AUTOINIT,
   VERSION,
   NT_LIBRARY,
   0,
   (char *) LibName,
   (char *) LibId,
   LibInitTable
};
/*FE*/

/* ---------------------------- implementation ---------------------------- */

/*FS*/ static Class *initClass(struct ClassBase *cb)
{
   Class *cl;

   if((cl = MakeClass(CLASSNAME,SUPERCLASSNAME,NULL,INSTANCESIZE,0)) != NULL)
   {
      DB(("dispatcher : %lx,size : %ld\n",dispatcher,INSTANCESIZE));

      cl->cl_Dispatcher.h_Entry = (HOOKFUNC) dispatcher;
      cl->cl_UserData = (ULONG) cb;

      AddClass(cl);
   }

   DB(("class : %lx\n",cl));

   return(cl);
}
/*FE*/
/*FS*/ static BOOL OpenLibraries(struct ClassBase *cb)
{
   BOOL rc = TRUE;
   /* system libraries must be opened correctly, because a pre 3.0 check
    * was done in LibInit()
    */

   IntuitionBase = OpenLibrary("intuition.library",39);
   DOSBase       = OpenLibrary("dos.library",      39);
   UtilityBase   = OpenLibrary("utility.library",  39);
#ifdef USE_GFXLIB
   GfxBase       = OpenLibrary("graphics.library", 39);
#endif

   if((DataTypesBase = OpenLibrary("datatypes.library",39)) == NULL)
      rc = FALSE;
#ifdef USE_IFFPARSELIB
   else
   {
      if((IFFParseBase  = OpenLibrary("iffparse.library", 39)) == NULL)
	 rc = FALSE;
   }
#endif

   return rc;
}
/*FE*/
/*FS*/ static void CloseLibraries(struct ClassBase *cb)
{
   CloseLibrary((struct Library *) UtilityBase);
   CloseLibrary((struct Library *) DOSBase);
   CloseLibrary((struct Library *) IntuitionBase);
   UtilityBase   = NULL;
   DOSBase       = NULL;
   IntuitionBase = NULL;
#ifdef USE_GFXLIB
   CloseLibrary((struct Library *) GfxBase);
   GfxBase       = NULL;
#endif

   CloseLibrary((struct Library *) DataTypesBase);
   DataTypesBase = NULL;

#ifdef USE_IFFPARSELIB
   CloseLibrary((struct Library *) IFFParseBase);
   IFFParseBase  = NULL;
#endif
}
/*FE*/

/*FS*/ LibCall Class *ObtainEngine (REGA6 struct ClassBase *cb)
{
   /* here we need no semaphore operations, because this function
    * can only be called between a OpenLibrary()/CloseLibrary()
    * call.
    */

   return cb->cb_Lib.cl_Class;
}
/*FE*/

/*FS*/ LibCall struct Library *LibInit (REGD0 struct ClassBase *cb, REGA0 BPTR seglist, REGA6 struct Library * sysbase)
{
   if(sysbase->lib_Version < 39)
      cb = NULL;
   else
   {
      cb->cb_SegList = seglist;
#ifdef USERCLASSBASEINIT
      cb->cb_UserInit = FALSE;
#endif
      cb->cb_SysBase = sysbase;
      cb->cb_Lib.cl_Lib.lib_Revision = REVISION;

      InitSemaphore(&cb->cb_Lock);
   }

   return (struct Library *) cb;
}
/*FE*/
/*FS*/LibCall LONG LibOpen (REGA6 struct ClassBase *cb)
{
   struct SignalSemaphore *lock= &cb->cb_Lock;
   LONG retval = (LONG) cb;

   ObtainSemaphore(lock);

   cb->cb_Lib.cl_Lib.lib_Flags &= ~LIBF_DELEXP;

   if(++cb->cb_Lib.cl_Lib.lib_OpenCnt == 1)
   {
      if(!OpenLibraries(cb))
	 retval = (LONG) NULL;
      else
      {
#ifdef SUPERCLASSLIBRARY
      if((cb->cb_SuperClassBase = OpenLibrary(SUPERCLASSLIBRARY,SUPERCLASSLIBRARYVERSION)) != NULL)
      {
#endif
	 cb->cb_Lib.cl_Class = initClass(cb);

#ifdef USERCLASSBASEINIT
	 if(!UserClassBaseInit(cb))
	 {
	    CloseLibraries(cb);
	    FreeClass(cb->cb_Lib.cl_Class);
	    cb->cb_Lib.cl_Class = NULL;
	 } else
	    cb->cb_UserInit = TRUE;
#endif

#ifdef SUPERCLASSLIBRARY
	 if(cb->cb_Lib.cl_Class == NULL)
	 {
	    CloseLibrary(SuperClassBase);
	    SuperClassBase = NULL;
	 }
      } else
      {
	 CloseLibraries(cb);
	 retval = (LONG) NULL;
      }
#endif
      }
   }

   if(retval == (LONG) NULL)
      cb->cb_Lib.cl_Lib.lib_OpenCnt--;

   ReleaseSemaphore(lock);

   DB(("return = 0x%lx\n",retval));

   return retval;
}
/*FE*/
/*FS*/LibCall LONG LibClose (REGA6 struct ClassBase *cb)
{
   LONG retval = NULL;

   DB(("cnt = %ld, class = 0x%lx\n",cb->cb_Lib.cl_Lib.lib_OpenCnt,cb->cb_Lib.cl_Class));

   if(cb->cb_Lib.cl_Lib.lib_OpenCnt > 0)
   {
      if(--cb->cb_Lib.cl_Lib.lib_OpenCnt == 0)
      {
	 ObtainSemaphore(&cb->cb_Lock);

	 if(cb->cb_Lib.cl_Class != NULL && FreeClass(cb->cb_Lib.cl_Class))
	    cb->cb_Lib.cl_Class = NULL;

#ifdef SUPERCLASSLIBRARY
	 if(cb->cb_Lib.cl_Class == NULL)
	 {
	    CloseLibrary(SuperClassBase);
	    SuperClassBase = NULL;
	 }
#endif

	 if(cb->cb_Lib.cl_Class == NULL)
	 {
#ifdef USERCLASSBASEINIT
	    UserClassBaseClean(cb);
#endif
	    CloseLibraries(cb);
	 }

	 ReleaseSemaphore(&cb->cb_Lock);
      }
   }

   DB(("cnt = %ld, class = 0x%lx\n",cb->cb_Lib.cl_Lib.lib_OpenCnt,cb->cb_Lib.cl_Class));

   if(cb->cb_Lib.cl_Lib.lib_Flags & LIBF_DELEXP)
      retval = LibExpunge (cb);

   DB(("return = 0x%lx\n",retval));

   return retval;
}
/*FE*/
/*FS*/LibCall LONG LibExpunge(REGA6 struct ClassBase *cb)
{
   BPTR seg = NULL;

   DB(("cnt = %ld, class = 0x%lx\n",cb->cb_Lib.cl_Lib.lib_OpenCnt,cb->cb_Lib.cl_Class));

   if(cb->cb_Lib.cl_Lib.lib_OpenCnt == 0)
   {
      seg = cb->cb_SegList;

      Remove((struct Node *) cb);

      FreeMem((APTR)((ULONG)(cb) - (ULONG)(cb->cb_Lib.cl_Lib.lib_NegSize)), cb->cb_Lib.cl_Lib.lib_NegSize + cb->cb_Lib.cl_Lib.lib_PosSize);
   } else
   {
      cb->cb_Lib.cl_Lib.lib_Flags |= LIBF_DELEXP;
   }

   DB(("return = 0x%lx\n",seg));

   return (LONG) seg;
}
/*FE*/
/*FS*/LibCall LONG LibReserved(REGA6 struct ClassBase *cb)
{
   return 0;
}
/*FE*/

