
/*
**
**  $VER: paranimdtcpatch.c 1.2 (12.3.97)
**  paranimdtcpatch 1.2
**
**  Patch animation.datatype / realtime.library to allow parallel running
**  animation.datatype's
**
**  (C) Copyright 1996/1997 by Roland 'Gizzy' Mainz
**          All Rights Reserved
**
*/


/* We did not want any Intuition defines before V36 */
#define INTUI_V36_NAMES_ONLY 1

/* amiga includes */
#include <exec/types.h>
#include <exec/execbase.h>
#include <exec/alerts.h>
#include <libraries/realtime.h>

/* amiga prototypes */
#include <clib/exec_protos.h>
#include <clib/utility_protos.h>
#include <clib/dos_protos.h>

/* amiga pragmas */
#include <pragmas/exec_pragmas.h>
#include <pragmas/utility_pragmas.h>
#include <pragmas/dos_pragmas.h>

/* ansi includes */
#include <string.h>

/* version string */
#include "paranimdtcpatch_rev.h"


/* SASC specific defines */
#define DISPATCHERFLAGS __saveds __asm
#define REGARGS __asm
#define REGD0 register __d0
/* ... */
#define REGA0 register __a0
#define REGA1 register __a1
#define REGA2 register __a2
/* ... */
#define REGA6 register __a6


/* misc defines */
#ifndef NAME
#define NAME "paranimdtcpatch"
#endif /* !NAME */

/* version_string */
STRPTR versionstring = VERSTAG;

/* shared libraries */
extern struct ExecBase *SysBase;
extern struct Library  *DOSBase;
       struct Library  *RealTimeBase,
                       *UtilityBase,
                       *DTClassBase;


REGARGS  struct Player *(*OldCreatePlayerA)( REGA0 struct TagItem *, REGA6 struct Library * ); /* original function vector */
volatile WORD            patch_in_use = 0UL; /* patch code in use counter */

/* prototypes */
static DISPATCHERFLAGS struct Player *NewCreatePlayerA( REGA0 struct TagItem * );
static BOOL                           AttemptRemoveFunction( struct Library *, LONG, APTR, APTR );


/****** paranimdtcpatch/paranimdtcpatch **************************************
*
*   NAME
*       paranimdtcpatch -- animation.datatype patch, allows multiple objects
*
*   FORMAT
*       paranimdtcpatch
*
*   TEMPLATE
*       ,
*
*   PATH
*       paranimdtcpatch:
*
*   PURPOSE
*       *** This is a HACK which may or may not affect other programs ***
*
*       Patch animation.datatype to allow multiple parallel running
*       objects of animation.datatype or it's subclasses.
*
*       During creation of mpeg#?.datatype (misc datatypes for system/video/
*       audiostreams) and SwitchWindow (object-oriented BOOPSI environment
*       (analyse, create or modify BOOPSI objects including datatypes ones)),
*       I wrote this patch to fix a bug in animation.datatype.
*
*   DESCRIPTION
*       This hack patches realtime.library/CreatePlayerA function.
*       All calls from a task named "Animation_ObjectHandler (Display)"
*       which attempt to use a conductor named "Main" will return
*       a player with an PRIVATE conductor (~0UL name, see
*       realtime.library/CreatePlayerA docs).
*       This fixes a bug in animation.datatype which avoids the usage of
*       multiple parallel running objects of animation.datatype or it's
*       subclasses.
*
*   KNWON BUGS
*       - HACK
*
*       - Does everytimes return success when exit.
*
*       animation.datatype V40.7 has no known serious bugs exept this one.
*       Mainly all other bugs while playing IFF ANIM-5 anims are caused by
*       stack promblems and the CBM anim.datatype.
*
*   HISTORY
*       V1.1
*           First release.
*
*       V1.2
*           - Recompiled with SAS/C 6.57
*
*           - Small code cleanup
*
*           - Moved the contents of the #?.guide file in this autodoc
*
*           - Modified to be compatible with animation.datatype V41, which
*             does not need this patch (simply disables itself).
*             It also locks animation.datatype in memory, avoiding that
*             another version may be come active without removing and
*             reinstalling this patch.
*
*   NOTE
*
*   INSTALLATION
*        After unpacking this archieve:
*
*        Shell:
*          - Unpack this archieve and copy the paranimdtcpatch to
*            C:
*        Copy CLONE FROM paranimdtcpatch TO C:
*
*        make/smake (SAS):
*          - Type 'smake install' in shell
*
*   USAGE
*        Put the following statement in your user-startup:
*
*        run <>NIL: paranimdtcpatch:paranimdtcpatch
*
*   AUTHOR's REQUEST
*        By  releasing  this program I do  not  place any obligations on you,
*        feel free to share this program with your  friends (and enemies).
*
*        If you want to blame me, report any bugs, or wants a new version
*        send your letter to:
*                        Roland Mainz
*                        Hohenstaufenstrae 8
*                        52388 Nrvenich
*                        GERMANY
*
*        Phone: (+49)(0)2426/4111
*
*        EMAIL is also available:
*        GISBURN@w-specht.rhein-ruhr.de
*
*        If you want to send me attachments larger than 1MB (up to 5MB,
*        more with my permission):
*        Up to March 1997 I'm reachable using this email address, too:
*        Reinhold.A.Mainz@KBV.DE
*
*        | Please put your name and address in your mails !
*        | German mailers should add their phone numbers.
*        | See BUGS section above when submitting bug reports.
*
*        Sorry, but I can only look once a week for mails.
*        If you don't hear something from me within three weeks, please
*        send your mail again (but watch about new releases) (problems with
*        this email port are caused by reconfigurations, hackers, network
*        problems etc.).
*
*        The  entire  "paranimdtcpatch"  package  may  be  noncommercially
*        redistributed, provided  that  the package  is always  distributed
*        in it's complete  form (including it's documentation).  A small copy
*        fee for media costs is okay but any kind of commercial distribution
*        is strictly forbidden! Comments  and  suggestions  how  to  improve
*        this program  are generally appreciated!
*
*        Thanks to David Junod, who wrote the animation.datatype and the
*        datatypes example code, Matt Dillon for his DICE, and Olaf 'Olsen'
*        Barthel for his help, ideas and some text clips from his
*        documentations.
*
*   SEE ALSO
*       animation.datatype, mpegvideo.datatype, mpegsystem.datatype
*       realtime.library/CreatePlayerA,
*       exec.library/SetFunction
*
*****************************************************************************
*
*/


long main( void )
{
    /* Obtain animation.datatype */
    if( DTClassBase = OpenLibrary( "datatypes/animation.datatype", 39UL ) )
    {
      /* animation.datatype V41 running ? */
      if( (DTClassBase -> lib_Version) < 41U )
      {
        if( UtilityBase = OpenLibrary( "utility.library", 39UL ) )
        {
          if( RealTimeBase = OpenLibrary( "realtime.library", 0UL ) )
          {
            OldCreatePlayerA = (REGARGS struct Player *(*)( REGA0 struct TagItem *, REGA6 struct Library * ))SetFunction( RealTimeBase, (-42L), (ULONG (*)())NewCreatePlayerA );

            /* Wait for quit */
            for( ;; )
            {
              Wait( SIGBREAKF_CTRL_C );

              if( AttemptRemoveFunction( RealTimeBase, (-42L), (APTR)OldCreatePlayerA, (APTR)NewCreatePlayerA ) )
              {
                if( patch_in_use )
                {
                  Printf( "patch node in use, try CTRL_C again\n" );

                  Delay( (TICKS_PER_SECOND / 2UL) );
                }
                else
                {
                  break;
                }
              }
              else
              {
                Printf( "can't restore original function entry\n" );
              }
            }

            CloseLibrary( RealTimeBase );
          }
          else
          {
            Printf( "can't open \"realtime.library\" vers 0\n" );
          }

          CloseLibrary( UtilityBase );
        }
        else
        {
          Printf( "can't open \"utility.library\" vers 39\n" );
        }
      }
      else
      {
        Printf( "animation.datatype V41 detected; this patch is not needed for this version\n" );
      }

      CloseLibrary( DTClassBase );
    }
    else
    {
      Printf( "can't open \"datatypes/animation.datatype\" vers 39\n" );
    }

    return( RETURN_OK );
}


static
DISPATCHERFLAGS
struct Player *NewCreatePlayerA( REGA0 struct TagItem *tags )
{
    struct Player *player = NULL;

    /* Be sure that we don't exit while someone is running inside this patch code */
    Forbid();

      patch_in_use++;

    Permit();

    for( ;; ) /* Used as a jump table (avoids a goto) */
    {
      /* The patch will only affect tasks which match all of the following conditions:
       * - The task name must be "Animation_ObjectHandler (Display)"
       * - A given PLAYER_Conductor tagitem
       * - The PLAYER_Conductor tagitem data value must point to a string "Main"
       */
      if( !strcmp( "Animation_ObjectHandler (Display)", (SysBase -> ThisTask -> tc_Node . ln_Name) ) )
      {
        struct TagItem *ti;

        if( ti = FindTagItem( PLAYER_Conductor, tags ) )
        {
          if( !strcmp( "Main", (STRPTR)(ti -> ti_Data) ) )
          {
            ULONG olddata; /* Saved old value from (ti -> ti_Data)
                            * The official way would be to copy the taglist, then modify the copy,
                            * otherwise this may affect "const"-marked data (BUG !!)
                            */

            olddata = ti -> ti_Data;

            ti -> ti_Data = ~0UL;    /* "Mark" conductor as private */

            player = (*OldCreatePlayerA)( tags, RealTimeBase ); /* Call original function */

            ti -> ti_Data = olddata; /* Restore old value */

            break; /* Leave jump table */
          }
        }
      }

      player = (*OldCreatePlayerA)( tags, RealTimeBase ); /* Call original function */

      break; /* End of jump table */
    }

    Forbid();

      patch_in_use--;

    Permit();

    return( player );
}


static
BOOL AttemptRemoveFunction( struct Library *lib, LONG offset, APTR oldfunc, APTR patchfunc )
{
    BOOL removed = FALSE;

    if( lib && offset && oldfunc && patchfunc )
    {
      APTR currfunc;

      Forbid();

       currfunc = (APTR)SetFunction( lib, offset, (ULONG (*)())oldfunc );

       /* Does the code returned from SetFunction match our patch code ? */
       if( currfunc == patchfunc )
       {
         removed = TRUE;
       }
       else
       {
         /* No, restore returned code */
         currfunc = (APTR)SetFunction( lib, offset, (ULONG (*)())currfunc );

         /* The following expression MUST never be TRUE */
         if( currfunc != oldfunc )
         {
           /* havoc found, crash expected */
           Alert( AN_Unknown + 0x1357 );
         }
       }

      Permit();
    }

    return( removed  );
}

