/*
 * Only some changes by Detlef Wuerkner, removed the V40 stuff and rewrote
 * DTS_ReadIntoBitMap43() since QRT is a different format :-)
 *
 *  The format is as follows:
 *
 *  (header:)
 *    wwww hhhh       - Width, Height (16 bits, LSB first)
 *
 *  (each scanline:)
 *    llll            - Line number (16 bits, LSB first)
 *    rr rr rr ...    - Red data for line (8 bits per pixel,
 *                       left to right, 0-255 (255=bright, 0=dark))
 *    gg gg gg ...    - Green data for line (8 bits per pixel,
 *                       left to right, 0-255 (255=bright, 0=dark))
 *    bb bb bb ...    - Blue data for line (8 bits per pixel,
 *                       left to right, 0-255 (255=bright, 0=dark))
 */

/*
**      $VER: libfuncs.c 43.6 (27.3.97)
**
**      datatype functions
**
**      (C) Copyright 1996-97 Andreas R. Kleinert
**      All Rights Reserved.
*/


ULONG __saveds __stdargs DTS_ReadIntoBitMap43(struct ClassBase *cb, Object * o, Class *cl, struct TagItem * attrs)
{
 struct BitMapHeader *bmhd;

 BOOL success = TRUE;

 ULONG width, height;

 UBYTE *readname, *buffer, *bufrgb;

 BPTR fh;

 UBYTE wstr[2], hstr[2], lstr[2];
 ULONG source_type;


 /* we can't handle clipboard, ram or other access modes */

 source_type = (ULONG) GetTagData (DTA_SourceType, DTST_FILE, attrs);
 if(source_type != DTST_FILE)
  {
   SetIoErr(ERROR_OBJECT_WRONG_TYPE);
   return(FALSE);
  }

 readname = (UBYTE *) GetTagData (DTA_Name, NULL, attrs);
 getdtattrs (cb, o, PDTA_BitMapHeader, &bmhd, TAG_DONE, NULL);


 fh = Open(readname, MODE_OLDFILE);
 if(!fh)
  {
   SetIoErr(ERROR_OBJECT_NOT_FOUND);
   return(FALSE);
  }

 /* get Width */
 if(!FRead(fh, &wstr[0], 2, 1))
  {
   Close(fh);
   return(FALSE);
  }

 /* get Height */
 if(!FRead(fh, &hstr[0], 2, 1))
  {
   Close(fh);
   return(FALSE);
  }

 width  = (wstr[1] << 8) + wstr[0];
 height = (hstr[1] << 8) + hstr[0];

 if((width==0)||(height==0))
  {
   Close(fh);
   return(FALSE);
  }

 bmhd->bmh_Width  = (bmhd->bmh_PageWidth  = width);
 bmhd->bmh_Height = (bmhd->bmh_PageHeight = height);
 bmhd->bmh_Depth = 24;

 setdtattrs(cb, o,
            DTA_ObjName,      readname,
            DTA_NominalHoriz, bmhd->bmh_Width,
            DTA_NominalVert,  bmhd->bmh_Height,
            PDTA_SourceMode,  MODE_V43,
            PDTA_ModeID,      DTS_GetBestModeID(bmhd->bmh_Width, bmhd->bmh_Height, 8),
            TAG_DONE);


 buffer = (APTR) AllocVec(width*3, MEMF_PUBLIC);
 bufrgb = (APTR) AllocVec(width*3, MEMF_PUBLIC);
 if(buffer && bufrgb)
  {
   UBYTE *bufdest, *bufred, *bufgreen, *bufblue;
   ULONG i,j;

   for(i=0; i<height; i++)
    {
     /* check linenum, read data */
     if(
         !FRead(fh, &lstr[0], 2, 1) ||
         (i != (lstr[1] << 8) + lstr[0]) ||
         !FRead(fh, buffer, width*3, 1)
       )
      {
       success = FALSE;
       SetIoErr (ERROR_SEEK_ERROR);
       break;
      }
     bufdest = bufrgb;
     bufred = buffer;
     bufgreen = bufred + width;
     bufblue =  bufgreen + width;
     for(j=0; j<width; j++)
      {
       *bufdest++=*bufred++;
       *bufdest++=*bufgreen++;
       *bufdest++=*bufblue++;
      }
     DoSuperMethod(cl, o, DTM_WRITEPIXELARRAY, bufrgb, RECTFMT_RGB, width*3, 0, i, width, 1);
    }
  }else
  {
   success = FALSE;
   SetIoErr (ERROR_NO_FREE_STORE);
  }
 if(buffer) FreeVec(buffer);
 if(bufrgb) FreeVec(bufrgb);
 Close(fh);
 return(success);
}
